from pathlib import Path

import typer
from bs4 import BeautifulSoup


def main(path: Path):
    for file in path.glob("*.html"):
        with open(file, "r") as f:
            soup = BeautifulSoup(f, "html.parser")

        modified_soup = modify_class_name(soup)

        with open(file, "w") as f:
            f.write(str(modified_soup))


def to_snake_case(s: str):
    return "".join(["_" + c.lower() if c.isupper() else c for c in str(s)]).strip("_")


def modify_class_name(soup: BeautifulSoup) -> BeautifulSoup:
    """Example: replace avatars.api.Datasets by avatars.ApiClient.datasets"""
    # Find all class entries in the docs
    all_classes = soup.find_all("dl", {"class": "py class"})
    if not all_classes:
        return soup

    # Extract the class/module signature from class entries starting with "avatars.api."
    client_and_api_class_definition_tags = [
        c.find("dt") for c in all_classes if "avatars.api." in c.find("dt").attrs["id"]
    ]

    # Get the module name tag from the signature
    module_name_tags = [
        c.find("span", {"class": "sig-prename descclassname"})
        for c in client_and_api_class_definition_tags
    ]
    module_name_tags = [c.find("span", {"class": "pre"}) for c in module_name_tags if c]

    # Get the class name tag from the signature
    class_name_tags = [
        c.find("span", {"class": "sig-name descname"})
        for c in client_and_api_class_definition_tags
    ]

    class_name_tags = [c.find("span", {"class": "pre"}) for c in class_name_tags if c]

    for module_name, class_name in zip(module_name_tags, class_name_tags):
        module_name.string = "avatars.ApiClient."
        class_name.string = to_snake_case(str(class_name.string))

    return soup


if __name__ == "__main__":
    typer.run(main)



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
# ---

# %% [markdown]
# # Tutorial 4: Time Series

# %% [markdown]
# In this tutorial, we will perform the avatarization of data containing time series variables. This is a specific case of a multitable avatarization

# %% [markdown]
# ## Connection

# %%
import os

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from avatar_yaml.models.parameters import ProjectionType
from avatar_yaml.models.schema import LinkMethod

from avatars.constants import PlotKind
from avatars.manager import Manager

username = os.environ.get("AVATAR_USERNAME", "")
password = os.environ.get("AVATAR_PASSWORD", "")

# %%
manager = Manager()  # or manager = Manager(base_url=https://your-server.com)
# Authenticate with the server
manager.authenticate(username, password)
# Verify that we can connect to the API server
manager.get_health()

# %% [markdown]
# ## Loading data

# %% [markdown]
# In this tutorial, we use data that contains readings from 3 sensors for 200 individuals. The data is divided into 3 datasets as follows:
# - `sensors_vanilla.csv`: demographic data on individuals, each line refers to one individual and contains 15 variables (similar to adult data seen in previous tutorials)
# - `sensors_timeseries1.csv`: time series data for `sensor1` and `sensor2`. The data contains an identifier variable (`id`) and a time variable (`t`). There are several lines for each individual.
# - `sensors_timeseries2.csv`: time series data for `sensor3`. Note that this time series data can have different timestamps than the other time series and a different number of measurements.

# %%
vanilla_df = pd.read_csv("../fixtures/sensors_vanilla.csv")
ts1_df = pd.read_csv("../fixtures/sensors_timeseries1.csv")
ts2_df = pd.read_csv("../fixtures/sensors_timeseries2.csv")

# %%
vanilla_df.head()

# %% [markdown]
# **Note that the vanilla data contains a column `id` required to link time series data point with each individual.**

# %%
ts1_df

# %%
ts2_df


# %% [markdown]
# ## Overview of data

# %% [markdown]
# We provide below a basic visualization function to better understand how the time series data look like.


# %%
def plot_series(
    df: pd.DataFrame,
    variable_to_plot: str,
    id_variable: str,
    time_variable: str,
    proportion_to_plot: float = 1.0,
    n_series_to_plot: int | None = None,
    figsize: tuple[int, int] = (14, 8),
) -> matplotlib.figure.Figure:
    """Plot given series."""
    if n_series_to_plot is None:
        n_series_to_plot = df[id_variable].unique().shape[0]

    df_tmp = df.copy()
    df_tmp = df_tmp.sort_values(by=[id_variable, time_variable]).reset_index(drop=True)

    cmap = plt.get_cmap("gist_rainbow")

    fig, ax = plt.subplots(figsize=figsize)
    for id_nb, id_name in enumerate(set(df_tmp[id_variable])):
        if id_nb > n_series_to_plot:
            break
        selected_records = df_tmp[df_tmp[id_variable] == id_name]
        n_points = int(proportion_to_plot * len(selected_records))
        selected_indices = np.linspace(0, len(selected_records) - 1, num=n_points)

        x = selected_records[time_variable].iloc[selected_indices]
        y = selected_records[variable_to_plot].iloc[selected_indices]
        ax.plot(x, y, color=cmap(id_nb / n_series_to_plot))
    ax.set_title(f"Series for variable {variable_to_plot}")
    return fig


# %% [markdown]
# and we use it to plot sensor data for some individuals

# %% [markdown]
# ### `sensor1` data

# %%
plot = plot_series(
    df := ts1_df,
    variable_to_plot="sensor1",
    id_variable="id",
    time_variable="t",
    proportion_to_plot=1.0,
    n_series_to_plot=None,
    figsize=(14, 8),
)

# %% [markdown]
# ###  `sensor2` data

# %%
plot = plot_series(
    df := ts1_df,
    variable_to_plot="sensor2",
    id_variable="id",
    time_variable="t",
    proportion_to_plot=1.0,
    n_series_to_plot=None,
    figsize=(14, 8),
)

# %% [markdown]
# ### `sensor3` data

# %%
plot = plot_series(
    df := ts2_df,
    variable_to_plot="sensor3",
    id_variable="id",
    time_variable="t",
    proportion_to_plot=1.0,
    n_series_to_plot=None,
    figsize=(14, 8),
)

# %% [markdown]
# Now that we know what data we are maniplating, we can anonymize it.

# %% [markdown]
# ## Avatars of time series data

# %% [markdown]
# Avatarizing time_series data is a specific case of multitable avatarization. Please see Tutorial3.
# You will need to use a specific LinkMethod : `TIME_SERIES` and to specify the time scale variable `t`

# %% [markdown]
# ### Upload data and save datasets

# %%
# First initialize the runner
runner = manager.create_runner("tutorial_time_series")

# Then upload the data
runner.add_table("vanilla", vanilla_df, individual_level=True, primary_key="id")
runner.add_table(
    "sensor1", ts1_df, primary_key="primary_key", foreign_keys=["id"], time_series_time="t"
)
runner.add_table(
    "sensor2", ts2_df, primary_key="primary_key", foreign_keys=["id"], time_series_time="t"
)

# %% [markdown]
# ### add links between tables

# %%
runner.add_link(
    parent_table_name="vanilla",
    parent_field="id",
    child_table_name="sensor1",
    child_field="id",
    method=LinkMethod.TIME_SERIES,
)

runner.add_link(
    parent_table_name="vanilla",
    parent_field="id",
    child_table_name="sensor2",
    child_field="id",
    method=LinkMethod.TIME_SERIES,
)

# %% [markdown]
# ### Create job and retrieve results
#
# This is done in a similar way than for the anonymization of classic tabular data.
#
# The attributes that are specific to time_series anonymization are:
# - `time_series_projection_type`: The projection parameters defines how the data will be transformed (aka projected) so that avatars can be generated.
# - `time_series_nf`:  This is the number of components used to represent each time series variable in the projection.
# - `time_series_nb_points` :
# - `time_series_method`:

# %%
runner.set_parameters("vanilla", k=5)
runner.set_parameters(
    "sensor1", time_series_projection_type=ProjectionType.FLATTEN, time_series_nf=10
)
runner.set_parameters(
    "sensor2", time_series_projection_type=ProjectionType.FLATTEN, time_series_nf=10
)

# %%
runner.run()
results = runner.get_all_results()

# %% [markdown]
# ### Retrieve avatars
#
# Once generated, the avatar dataset are stored on the server and can be retrieved using their dataset names.

# %% [markdown]
# ### Checking your avatars: vanilla data

# %%
avatars_vanilla = runner.shuffled("vanilla")
avatars_vanilla.head()

# %% [markdown]
# ### Checking your avatars: `sensor1` and `sensor2`

# %%
avatars_ts1_df = runner.shuffled("sensor1")
avatars_ts1_df.head()

# %%
# If you are having issues rendering the plot in your notebook, you can use the following line to open the plot in your browser
# runner.render_plot("sensor1", plot_kind=PlotKind.RAW_SERIES, open_in_browser=True)
runner.render_plot("sensor1", plot_kind=PlotKind.RAW_SERIES)

# %%
# If you are having issues rendering the plot in your notebook, you can use the following line to open the plot in your browser
# runner.render_plot("sensor2", plot_kind=PlotKind.RAW_SERIES, open_in_browser=True)
runner.render_plot("sensor2", plot_kind=PlotKind.RAW_SERIES)

# %% [markdown]
# ### Checking your avatars: `sensor3`

# %%
avatars_ts2_df = runner.shuffled("sensor2")
avatars_ts2_df.head()

# %%
# If you are having issues rendering the plot in your notebook, you can use the following line to open the plot in your browser
# runner.render_plot("sensor2", plot_kind=PlotKind.RAW_SERIES, open_in_browser=True)
runner.render_plot("sensor2", plot_kind=PlotKind.RAW_SERIES)

# %% [markdown]
# ## Privacy metrics for time series

# %% [markdown]
# As for any data, privacy metrics need to be computed on the output time series data to confirm that it is anonymous. We first retrieve the data required by the privacy metrics, that is the original datasets and their sensitive unshuffled counterparts. The unshuffled datasets are only used for computing the privacy metrics

# %% [markdown]
# Metric results are calculated for each dataset and are stored in `privacy_job.result`

# %%
for metric_details in runner.privacy_metrics("vanilla")[0].items():
    print(metric_details)

# %%
for metric_details in runner.privacy_metrics("sensor1")[0].items():
    print(metric_details)

# %%
for metric_details in runner.privacy_metrics("sensor2")[0].items():
    print(metric_details)

# %% [markdown]
# ## Signal metrics for time series data

# %%
for metric_details in runner.signal_metrics("vanilla")[0].items():
    print(metric_details)

# %%
for metric_details in runner.signal_metrics("sensor1")[0].items():
    print(metric_details)

# %%
for metric_details in runner.signal_metrics("sensor2")[0].items():
    print(metric_details)

# %% [markdown]
# ## Download report

# %%
runner.download_report("time_series_report.pdf")



import pytest
from pydantic import HttpUrl

from avatars.config import Config


@pytest.fixture
def clean_env(monkeypatch: pytest.MonkeyPatch) -> None:
    """Clear all AVATAR-related environment variables."""
    env_vars = [
        "AVATAR_BASE_URL",
        "AVATAR_BASE_API_URL",
        "AVATAR_STORAGE_ENDPOINT_URL",
        "STORAGE_ENDPOINT_URL",
    ]
    for var in env_vars:
        monkeypatch.delenv(var, raising=False)


class TestConfig:
    """Tests for the Config class."""


class TestVerifyCompatibility:
    def test_default_verify_compatibility(self) -> None:
        """Test that VERIFY_COMPATIBILITY defaults to True."""
        config = Config()
        assert config.VERIFY_COMPATIBILITY is True

    def test_verify_compatibility_from_env_true(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test that VERIFY_COMPATIBILITY can be set to True via env var."""
        monkeypatch.setenv("AVATAR_VERIFY_COMPATIBILITY", "true")
        config = Config()
        assert config.VERIFY_COMPATIBILITY is True

    def test_verify_compatibility_from_env_false(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test that VERIFY_COMPATIBILITY can be set to False via env var."""
        monkeypatch.setenv("AVATAR_VERIFY_COMPATIBILITY", "false")
        config = Config()
        assert config.VERIFY_COMPATIBILITY is False


class TestStorageEndpointURL:
    """Tests for STORAGE_ENDPOINT_URL configuration."""

    def test_default_storage_endpoint_url(self, clean_env: None) -> None:
        """Test that STORAGE_ENDPOINT_URL defaults to the SaaS URL."""
        config = Config()
        assert str(config.STORAGE_ENDPOINT_URL) == "https://www.octopize.app/storage"

    @pytest.mark.parametrize(
        "env_name",
        [
            "AVATAR_STORAGE_ENDPOINT_URL",
            "STORAGE_ENDPOINT_URL",  # backwards compatibility
        ],
    )
    def test_storage_endpoint_url_from_env(
        self, env_name: str, clean_env: None, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Test that STORAGE_ENDPOINT_URL can be set via env var multiple ways."""
        test_url = "https://storage.example.com"
        monkeypatch.setenv(env_name, test_url)
        config = Config()
        # HttpUrl adds trailing slash
        assert str(config.STORAGE_ENDPOINT_URL) == test_url + "/"


class TestBaseURLConstruction:
    """Tests for BASE_URL construction of API and storage URLs."""

    def test_base_url_from_env(self, clean_env: None, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test that BASE_URL can be set via environment variable."""
        monkeypatch.setenv("AVATAR_BASE_URL", "https://env.example.com")
        config = Config()
        assert str(config.BASE_API_URL) == "https://env.example.com/api"
        assert str(config.STORAGE_ENDPOINT_URL) == "https://env.example.com/storage"

    def test_explicit_base_api_url_without_base_url(self, clean_env: None) -> None:
        """Test that BASE_API_URL can be set explicitly without BASE_URL."""
        config = Config(BASE_API_URL=HttpUrl("https://custom-api.example.com"))
        assert str(config.BASE_API_URL) == "https://custom-api.example.com/"
        assert str(config.STORAGE_ENDPOINT_URL) == "https://www.octopize.app/storage"

    def test_explicit_storage_url_without_base_url(self, clean_env: None) -> None:
        """Test that STORAGE_ENDPOINT_URL can be set explicitly without BASE_URL."""
        config = Config(STORAGE_ENDPOINT_URL=HttpUrl("https://custom-storage.example.com"))
        assert str(config.BASE_API_URL) == "https://www.octopize.app/api"
        assert str(config.STORAGE_ENDPOINT_URL) == "https://custom-storage.example.com/"

    def test_explicit_both_urls_without_base_url(self, clean_env: None) -> None:
        """Test that both URLs can be set explicitly without BASE_URL."""
        config = Config(
            BASE_API_URL=HttpUrl("https://custom-api.example.com"),
            STORAGE_ENDPOINT_URL=HttpUrl("https://custom-storage.example.com"),
        )
        assert str(config.BASE_API_URL) == "https://custom-api.example.com/"
        assert str(config.STORAGE_ENDPOINT_URL) == "https://custom-storage.example.com/"

    def test_base_url_with_base_api_url_raises_error(self, clean_env: None) -> None:
        """Test that setting both BASE_URL and BASE_API_URL raises ValueError."""
        with pytest.raises(ValueError, match="Cannot set BASE_URL together with"):
            Config(
                BASE_URL=HttpUrl("https://example.com"),
                BASE_API_URL=HttpUrl("https://api.example.com"),
            )

    def test_base_url_with_storage_url_raises_error(self, clean_env: None) -> None:
        """Test that setting both BASE_URL and STORAGE_ENDPOINT_URL raises ValueError."""
        with pytest.raises(ValueError, match="Cannot set BASE_URL together with"):
            Config(
                BASE_URL=HttpUrl("https://example.com"),
                STORAGE_ENDPOINT_URL=HttpUrl("https://storage.example.com"),
            )

    def test_base_url_with_both_urls_raises_error(self, clean_env: None) -> None:
        """Test that setting BASE_URL with both other URLs raises ValueError."""
        with pytest.raises(ValueError, match="Cannot set BASE_URL together with"):
            Config(
                BASE_URL=HttpUrl("https://example.com"),
                BASE_API_URL=HttpUrl("https://api.example.com"),
                STORAGE_ENDPOINT_URL=HttpUrl("https://storage.example.com"),
            )



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
# ---

# %% [markdown]
# # Tutorial 6: Data Augmentation
#
# In this tutorial, you'll learn how to use Avatar's data augmentation capabilities to address class imbalance and increase dataset size while preserving statistical properties and privacy.

# %% [markdown]
# ## Connection

# %%
import os

import pandas as pd
from avatar_yaml.models.parameters import AugmentationStrategy

from avatars.constants import ColumnType, PlotKind
from avatars.manager import Manager
from avatars.models import JobKind

base_url = os.environ.get("AVATAR_BASE_URL", "http://localhost:8080/api")
username = os.environ.get("AVATAR_USERNAME", "user_integration")
password = os.environ.get("AVATAR_PASSWORD", "password_integration")

# %%
manager = Manager()  # or manager = Manager(base_url=https://your-server.com)
# Authenticate with the server
manager.authenticate(username, password)

# %% [markdown]
# ## What is Data Augmentation with Avatar?
#
# Data augmentation uses avatarization to generate synthetic individuals that share the same statistical properties as your original dataset. This is particularly useful when:
#
# - **Your dataset has class imbalance**: Some categories have too few samples for effective model training
# - **You need more data**: Your dataset is too small and you want to increase its size while maintaining privacy
#
# Avatar offers two main augmentation strategies:
#
# 1. **Class Balancing**: Increase the number of individuals in underrepresented classes of a categorical variable
# 2. **Multiplier Augmentation**: Apply a multiplication factor to increase the entire dataset size uniformly

# %% [markdown]
# ## Example 1: Balancing Minority Classes
#
# Let's use the Wisconsin Breast Cancer Diagnostic (WBCD) dataset, which has an imbalanced class distribution. We'll balance the second minority class to improve model training.
#
# ### Step 1: Load and Explore the Data

# %%
df = pd.read_csv("../fixtures/wbcd.csv")

# Check the class distribution
print("Original class distribution:")
print(df["Class"].value_counts())
print(f"\nTotal samples: {len(df)}")

# %%
df.sample(5)

# %% [markdown]
# ### Step 2: Configure the Augmentation
#
# We'll use the `minority` strategy, which increases the minority class to match the majority class size.

# %%
# Create a runner for this augmentation job
runner = manager.create_runner("data_augmentation_minority")

dtypes = {}
for col in df.columns:
    dtypes[col] = ColumnType.CATEGORY
# Upload the data and specify thats all variables are categorical column
runner.add_table("wbcd", df, types=dtypes)

# Configure augmentation parameters
runner.set_parameters(
    table_name="wbcd",
    k=5,
    data_augmentation_strategy=AugmentationStrategy.minority,  # Balance minority class
    data_augmentation_target_column="Class",  # Column to balance
)

# %% [markdown]
# ### Step 3: Run the Augmentation

# %%
# Run the avatarization with augmentation
runner.run(jobs_to_run=[JobKind.standard])

# %% [markdown]
# ### Step 4: Retrieve and Verify the Augmented Data

# %%
# Retrieve the augmented dataset
augmented_data = runner.shuffled("wbcd")

# Check the new class distribution
print("Augmented class distribution:")
print(augmented_data["Class"].value_counts())
print(f"\nTotal samples: {len(augmented_data)}")
print(f"Original samples: {len(df)}")
print(f"New samples generated: {len(augmented_data) - len(df)}")

# %% [markdown]
# #### Understanding the Results
#
# Looking at the output above, we can observe:
#
# 1. **Class Balance Achieved**: The minority class (4) has been augmented  to match the majority class
# 2. **Synthetic Records Generated**: Avatar created new synthetic records to balance the dataset
# 3. **Near-Perfect Balance**: The classes are now balanced with approximately equal counts
#
# **Why not exactly the same count?**
#
# The slight variation occurs because:
# - Avatarization is inherently a **stochastic process** that introduces controlled randomness
# - This randomness is a key feature that ensures **privacy preservation** -
# - The difference is minimal (less than 1%) and doesn't affect the practical benefit of class balancing

# %% [markdown]
# ### Step 5: Visualize the Results
#
# You can generate plots to visualize the distribution and verify the augmentation quality.

# %%
# View the distribution plot showing the balanced classes
runner.render_plot("wbcd", PlotKind.DISTRIBUTION, open_in_browser=False)

# %%
# Visualize the class separation in 2D space
runner.render_plot("wbcd", PlotKind.CLASS_PROJECTION_2D, open_in_browser=False)

# %% [markdown]
# ## Example 2: Multiplier Augmentation
#
# Multiplier augmentation increases your entire dataset by a specified factor, regardless of class distribution. This is useful when you simply need more data for training.
#
# ### Use cases:
# - Your dataset is too small for effective model training
# - You want to test model performance with larger datasets
# - You need to generate synthetic data while preserving privacy

# %%
runner_augmented_factorial = manager.create_runner("data_augmentation_factorial")

runner_augmented_factorial.add_table("wbcd", df, types=dtypes)

# Configure to balance all classes to the majority
runner_augmented_factorial.set_parameters(
    table_name="wbcd",
    k=5,
    data_augmentation_strategy=3,  # Multiplier augmentation
)

# Run the augmentation
runner_augmented_factorial.run(jobs_to_run=[JobKind.standard])

# Retrieve and verify the augmented dataset
shuffle = runner_augmented_factorial.shuffled("wbcd")

print(f"Original dataset size: {len(df)}")
print(f"Augmented dataset size: {len(shuffle)}")
print(f"Multiplication factor: {len(shuffle) / len(df):.2f} x")

# Verify class proportions are preserved
print("\nOriginal class proportions:")
print(df["Class"].value_counts(normalize=True))
print("\nAugmented class proportions:")
print(shuffle["Class"].value_counts(normalize=True))

# %% [markdown]
# #### Understanding Multiplier Augmentation Results
#
# The output demonstrates how multiplier augmentation maintains dataset characteristics while scaling size:
#
# 1. **Exact Multiplication**: The dataset grew from 683 to 2,049 samples, achieving precisely the 3.00x factor requested
# 2. **Preserved Class Proportions**: The class distribution remained nearly identical:
# 3. **Statistical Consistency**: The tiny variations (0.1-0.2%) confirm that anonymization preserves the original data's statistical properties while generating synthetic records
#
# Unlike class balancing strategies, multiplier augmentation **maintains existing imbalances** - it's useful when you need more data but want to preserve the natural class distribution of your dataset.

# %% [markdown]
# ## Summary of Augmentation Strategies
#
# | Strategy | Purpose | When to Use | Parameter |
# |----------|---------|-------------|-----------|
# | `minority` | Balance minority class to match majority | Binary classification with imbalance |  `data_augmentation_strategy` &  `data_augmentation_target_column` |
# | `not_majority` | Balance all classes to match the largest | Multi-class with multiple underrepresented classes |  `data_augmentation_strategy` &  `data_augmentation_target_column` |
# | `multiplier` | Multiply entire dataset size | Need more data while preserving proportions | `data_augmentation_strategy` |
# | `custom proportion` | Apply custom proportion to each class | Custom project | `data_augmentation_strategy` & `data_augmentation_target_column` |
#
# ### Key Points:
# - All strategies maintain the statistical properties of your original data
# - Augmented data is synthetic and privacy-preserving
# - You can adjust the `k` parameter to control privacy (higher k = more privacy, less utility)
# - Verify the augmented data distribution matches your expectations

# %% [markdown]
# ## Data Augmentation with Full Pipeline
#
# You can also generate privacy metrics, signal metrics, and reports alongside data augmentation. The workflow remains the same as in previous tutorials.

# %%
runner_full_process = manager.create_runner("data_augmentation_full_process")

runner_full_process.add_table("wbcd", df, types=dtypes)

# Configure to balance all classes to the majority
runner_full_process.set_parameters(
    table_name="wbcd",
    k=5,
    data_augmentation_strategy=AugmentationStrategy.not_majority,
    data_augmentation_target_column="Mitoses",
)

# Run the augmentation
runner_full_process.run()  # Avatarization then privacy metrics, signal metrics, and report

# Retrieve and verify the augmented dataset
shuffled = runner_full_process.shuffled("wbcd")

# %% [markdown]
# We compute metrics on multiple attack scenarios. Here is a table summary of the computation types:
#
#
# | Metric Type                    | Original Data            | Avatar Data                | Privacy Answer                                                                                        |
# | ------------------------------ | ------------------------ | -------------------------- | ----------------------------------------------------------------------------------------------------- |
# | **STANDALONE**                 | All originals            | All avatars                | Can any avatar record compromise the privacy of its corresponding original record?                    |
# | **BARYCENTERED**               | All originals            | Barycenters                | Can the aggregation of all avatars for an original record compromise its privacy?                     |
# | **CLASS_BALANCING_BARYCENTER** | Only augmented originals | Only augmented barycenters | Does the augmentation process increase privacy risk for originals that were augmented multiple times? |
#

# %%
# Retrieve privacy metrics
print("*** Privacy metrics ***")
for method in runner_full_process.privacy_metrics("wbcd"):
    print(f"Computation type: {method['metadata']['computation_type']}")
    print(f"  Hidden rate: {method['hidden_rate']}")
    print(f"  Distance to closest: {method['distance_to_closest']}")

# %%
runner_full_process.download_report("wbcd_report.pdf")  # report with specific augmentation metrics



__version__ = "1.12.0"

# Re-export main public entry points for convenience so users can do for example:
# from avatars import Manager, Runner, ApiClient
# Auto-install the crash handler for enhanced error reporting
from avatars import crash_handler as _crash_handler  # noqa: F401
from avatars.client import ApiClient  # noqa: F401
from avatars.manager import Manager  # noqa: F401
from avatars.runner import Runner  # noqa: F401

__all__ = [
    "__version__",
    "Manager",
    "ApiClient",
    "Runner",
]



from pydantic import AliasChoices, Field, HttpUrl, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self

UNCONFIGURED = "unconfigured"

DEFAULT_API_URL = "https://www.octopize.app/api"
DEFAULT_STORAGE_URL = "https://www.octopize.app/storage"

DEFAULT_CRASH_REPORT_PATH = "crash_report.txt"


def construct_urls(
    base_url: HttpUrl | None = None,
    base_api_url: HttpUrl | None = None,
    storage_endpoint_url: HttpUrl | None = None,
    use_defaults: bool = True,
) -> tuple[HttpUrl | None, HttpUrl | None]:
    """Construct API and storage URLs from configuration values.

    Ensures that either BASE_URL is used to construct both URLs,
    or BASE_API_URL and STORAGE_ENDPOINT_URL are explicitly set.
    Mixing the two patterns is not allowed.

    Parameters
    ----------
    base_url : HttpUrl | None
        The base URL to construct API and storage URLs from.
    base_api_url : HttpUrl | None
        The explicit API URL.
    storage_endpoint_url : HttpUrl | None
        The explicit storage endpoint URL.

    Returns
    -------
        A tuple of (base_api_url, storage_endpoint_url).

    Raises
    ------
    ValueError
        If BASE_URL is set together with BASE_API_URL or STORAGE_ENDPOINT_URL.
    """
    base_url_set = base_url is not None
    api_url_set = base_api_url is not None
    storage_url_set = storage_endpoint_url is not None

    # Case 1: BASE_URL is set, construct the other URLs
    if base_url_set:
        if api_url_set or storage_url_set:
            raise ValueError(
                "Cannot set BASE_URL together with BASE_API_URL or "
                "STORAGE_ENDPOINT_URL. Either use BASE_URL alone, or set "
                "BASE_API_URL and STORAGE_ENDPOINT_URL explicitly."
            )
        # Construct URLs from BASE_URL
        base = str(base_url).rstrip("/")
        return HttpUrl(f"{base}/api"), HttpUrl(f"{base}/storage")

    # Case 2: Individual URLs are set (or use defaults)
    if use_defaults:
        base_api_url = base_api_url or HttpUrl(DEFAULT_API_URL)
        storage_endpoint_url = storage_endpoint_url or HttpUrl(DEFAULT_STORAGE_URL)

    return base_api_url, storage_endpoint_url


class Config(BaseSettings):
    """Configuration settings for the Avatar client.

    This configuration class supports automatic loading from environment variables.
    When both environment variables and explicit Config object values are set,
    the explicitly provided values take precedence.

    Attributes
    ----------
    BASE_API_URL : str
        The base URL of the Avatar API server.
        Default is "http://www.octopize.app/api" for SaaS server.
    TIMEOUT : int
        Timeout in seconds for API requests.
        Default is 5 seconds.
    SHOULD_VERIFY_SSL : bool
        Whether to verify SSL certificates on the server.
        Default is True.
    STORAGE_ENDPOINT_URL : str
        The storage endpoint URL for file uploads/downloads.
        Default is "http://www.octopize.app/storage" for SaaS server.
    VERIFY_COMPATIBILITY : bool
        Whether to verify client-server compatibility on authentication.
        Default is True.
    API_KEY : str | None
        Optional API key for authentication. When provided, this will be used
        instead of username/password authentication. Can be set via the
        AVATAR_API_KEY environment variable.

    Notes
    -----

    Add the AVATAR_ prefix to environment variables to set these values from the environment.

    To override configuration values in tests or code, pass them as keyword arguments:
        config = Config(BASE_API_URL="https://test.com", VERIFY_COMPATIBILITY=False)
    """

    model_config = SettingsConfigDict(
        extra="ignore",
        env_prefix="AVATAR_",
    )
    BASE_URL: HttpUrl | None = None
    BASE_API_URL: HttpUrl | None = None
    TIMEOUT: int = 60
    SHOULD_VERIFY_SSL: bool = True
    STORAGE_ENDPOINT_URL: HttpUrl | None = Field(
        default=None,
        # Allow the environment variable to be set without prefix for backward compatibility
        validation_alias=AliasChoices("STORAGE_ENDPOINT_URL", "AVATAR_STORAGE_ENDPOINT_URL"),
    )
    VERIFY_COMPATIBILITY: bool = True
    API_KEY: str | None = None
    CRASH_REPORT_PATH: str = DEFAULT_CRASH_REPORT_PATH

    @model_validator(mode="after")
    def construct_urls_from_base(self) -> Self:
        """Construct API and storage URLs based on BASE_URL if provided.

        This validator runs after model initialization to ensure that
        BASE_API_URL and STORAGE_ENDPOINT_URL are correctly set based
        on BASE_URL if it is provided.

        Returns
        -------
        Self
            The validated Config instance with constructed URLs.
        """
        api_url, storage_url = construct_urls(
            base_url=self.BASE_URL,
            base_api_url=self.BASE_API_URL,
            storage_endpoint_url=self.STORAGE_ENDPOINT_URL,
        )
        self.BASE_API_URL = api_url
        self.STORAGE_ENDPOINT_URL = storage_url
        return self


def get_config() -> Config:
    return Config()


config: Config = get_config()



import numpy as np
import pandas as pd
import pandas.testing as pd_testing
import pytest

from avatars.processors import RelativeDifferenceProcessor

ORIGINAL_DF = pd.DataFrame(
    {
        "variable_1": [100, 150, 120, 100],
        "variable_5": [110, 180, 130, 100],
        "variable_6": [210, 340, 280, 220],
        "variable_7": [110, 180, 130, np.nan],
    }
)
TRANSFORMED_DF = pd.DataFrame(
    {
        "variable_1": [100, 150, 120, 100],
        "variable_5": [10, 30, 10, 0],
        "variable_6": [0, 10, 30, 20],
        "variable_7": [110, 180, 130, np.nan],
    }
)


def test_preprocess_one_variable() -> None:
    # Check that one variable is transformed into a relative difference to another one.
    processor = RelativeDifferenceProcessor(
        target="variable_5",
        references=["variable_1"],
    )
    df = processor.preprocess(df=ORIGINAL_DF)

    pd_testing.assert_frame_equal(
        TRANSFORMED_DF[["variable_1", "variable_5"]],
        df[["variable_1", "variable_5"]],
        check_dtype=False,
    )


def test_preprocess_many_variables() -> None:
    # Check that one variable is transformed into a relative difference to a group
    # of other variables.
    processor = RelativeDifferenceProcessor(
        target="variable_6",
        references=["variable_1", "variable_5"],
    )
    df = processor.preprocess(df=ORIGINAL_DF)

    pd_testing.assert_frame_equal(
        TRANSFORMED_DF[["variable_6"]],
        df[["variable_6"]],
        check_dtype=False,
    )


def test_postprocess_one_variable() -> None:
    # Check that one variable transformed into a relative difference to another one can
    # be re-expressed in its original form.
    processor = RelativeDifferenceProcessor(target="variable_5", references=["variable_1"])
    df = processor.postprocess(source=pd.DataFrame(), dest=TRANSFORMED_DF)

    pd_testing.assert_frame_equal(
        ORIGINAL_DF[["variable_1", "variable_5"]],
        df[["variable_1", "variable_5"]],
    )


def test_postprocess_many_variables() -> None:
    # Check that one variable is transformed into a relative difference to a group of
    # other variables can be re-expressed in its original form.
    processor = RelativeDifferenceProcessor(
        target="variable_6", references=["variable_1", "variable_5"]
    )
    pre_df = processor.preprocess(df=ORIGINAL_DF)
    post_df = processor.postprocess(source=pd.DataFrame(), dest=pre_df)

    pd_testing.assert_frame_equal(
        ORIGINAL_DF[["variable_1", "variable_5", "variable_6"]],
        post_df[["variable_1", "variable_5", "variable_6"]],
        check_dtype=False,
    )


def test_preprocess_postprocess_many_variables() -> None:
    processor = RelativeDifferenceProcessor(
        target="variable_6", references=["variable_1", "variable_5"]
    )
    pre_df = processor.preprocess(df=ORIGINAL_DF)
    post_df = processor.postprocess(source=ORIGINAL_DF, dest=pre_df)

    pd_testing.assert_frame_equal(
        ORIGINAL_DF,
        post_df,
        check_dtype=False,
    )


def test_preprocess_raises_when_nan_in_references() -> None:
    processor = RelativeDifferenceProcessor(
        target="variable_1",
        references=["variable_7"],
    )
    with pytest.raises(ValueError, match="Expected no missing values for `references`"):
        processor.preprocess(df=ORIGINAL_DF)


def test_preprocess_with_rename_nodrop() -> None:
    """Verify that a new column is created and original target column kept."""
    processor = RelativeDifferenceProcessor(
        target="variable_5",
        references=["variable_1"],
        target_rename="variable_5_renamed",
    )
    df = processor.preprocess(df=ORIGINAL_DF)
    assert "variable_5" in df.columns and "variable_5_renamed" in df.columns


def test_preprocess_with_rename_drop() -> None:
    """Verify that a new column is created and original target column dropped."""
    processor = RelativeDifferenceProcessor(
        target="variable_5",
        references=["variable_1"],
        target_rename="variable_5_renamed",
        drop_original_target=True,
    )
    df = processor.preprocess(df=ORIGINAL_DF)
    assert "variable_5" not in df.columns and "variable_5_renamed" in df.columns


@pytest.mark.parametrize("drop_original_target", [True, False])
def test_postprocess_with_rename(drop_original_target: bool) -> None:
    """Verify that the transformed and renamed variable can be inverse transformed."""
    processor = RelativeDifferenceProcessor(
        target="variable_5",
        references=["variable_1"],
        target_rename="variable_5_renamed",
        drop_original_target=drop_original_target,
    )
    df = processor.preprocess(df=ORIGINAL_DF)
    df = processor.postprocess(source=pd.DataFrame(), dest=df)

    pd_testing.assert_frame_equal(
        ORIGINAL_DF[["variable_1", "variable_5"]],
        df[["variable_1", "variable_5"]],
        check_dtype=False,
    )


@pytest.mark.parametrize("scaling_unit", [2, 10])
def test_preprocess_with_scaling_unit(scaling_unit: int) -> None:
    """Verify that scaling_unit_parameter is working for preprocess."""
    processor = RelativeDifferenceProcessor(
        target="variable_5",
        references=["variable_1"],
        scaling_unit=scaling_unit,
    )
    df = processor.preprocess(df=ORIGINAL_DF)

    pd_testing.assert_frame_equal(
        TRANSFORMED_DF[["variable_5"]] / scaling_unit,
        df[["variable_5"]],
        check_dtype=False,
    )


@pytest.mark.parametrize("scaling_unit", [2, 10])
def test_postprocess_with_scaling_unit(scaling_unit: int) -> None:
    """Verify that scaling_unit_parameter is working for postprocess."""
    processor = RelativeDifferenceProcessor(
        target="variable_5",
        references=["variable_1"],
        scaling_unit=scaling_unit,
    )
    df = processor.preprocess(df=ORIGINAL_DF)
    df = processor.postprocess(source=pd.DataFrame(), dest=df)

    pd_testing.assert_frame_equal(df, ORIGINAL_DF, check_dtype=False)


def test_preprocess_wrong_parameters() -> None:
    with pytest.raises(
        ValueError,
        match="Expected drop_original_target to be False if a target_rename is None*",
    ):
        RelativeDifferenceProcessor(
            target="variable_5",
            references=["variable_1"],
            target_rename=None,
            drop_original_target=True,
        )


def test_preprocess_wrong_references_variables() -> None:
    processor = RelativeDifferenceProcessor(
        target="variable_5",
        references=["wrong_variable"],
        target_rename="None",
    )

    with pytest.raises(
        ValueError, match="Expected all reference variables in dataset columns, got *"
    ):
        processor.preprocess(df=ORIGINAL_DF)


def test_postprocess_wrong_references_variables() -> None:
    processor = RelativeDifferenceProcessor(
        target="variable_5",
        references=["wrong_variable"],
        target_rename="None",
    )

    with pytest.raises(
        ValueError, match="Expected all reference variables in dataset columns, got *"
    ):
        processor.postprocess(ORIGINAL_DF, ORIGINAL_DF)



"""Tests for JobLauncher class."""

from uuid import uuid4

import pytest
from avatar_yaml import Config

from avatars.job_launcher import JobLauncher
from avatars.models import JobKind
from tests.unit.conftest import FakeApiClient


class TestJobLauncher:
    """Test suite for JobLauncher class."""

    @pytest.fixture(scope="function")
    def config(self):
        """Create a fresh config with a table for each test."""
        config = Config(set_name="test_set")
        config.create_table(
            table_name="test_table",
            original_volume="volume",
            original_file="test.csv",
        )
        return config

    @pytest.fixture(scope="function")
    def launcher(self, config) -> JobLauncher:
        """Create a fresh JobLauncher instance with fake API client for each test."""
        client = FakeApiClient(tables=["test_table"])
        return JobLauncher(client=client, config=config)

    @pytest.mark.parametrize(
        "job_kind,pia_report,expected",
        [
            (JobKind.standard, False, "standard"),
            (JobKind.standard, True, "standard"),
            (JobKind.privacy_metrics, False, "privacy_metrics"),
            (JobKind.privacy_metrics, True, "privacy_metrics"),
            (JobKind.signal_metrics, False, "signal_metrics"),
            (JobKind.signal_metrics, True, "signal_metrics"),
            (JobKind.report, False, "report"),
            (JobKind.report, True, "report_pia"),
        ],
    )
    def test_get_parameters_name(self, launcher, job_kind, pia_report, expected):
        """Test get_parameters_name for different job kinds."""
        assert launcher.get_parameters_name(job_kind, pia_report=pia_report) == expected

    def test_launch_standard_job_without_parameters_raises_error(self, launcher):
        """Test that launching standard job without avatarization parameters raises error."""
        set_name = str(uuid4())
        with pytest.raises(
            ValueError,
            match="Expected k or epsilon to be set to run an avatarization job,",
        ):
            launcher.launch_job(JobKind.standard, set_name)

    def test_launch_standard_job_with_avatarization_parameters(self, launcher):
        """Test launching standard job with avatarization parameters."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)

        assert launcher.set_name == set_name
        assert "standard" in launcher.jobs.keys()
        assert launcher.has_job(JobKind.standard)

    def test_launch_standard_job_with_dp_parameters(self, launcher):
        """Test launching standard job with differential privacy parameters."""
        launcher.config.create_avatarization_dp_parameters("test_table", epsilon=1.0, ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)

        assert launcher.set_name == set_name
        assert "standard" in launcher.jobs
        assert launcher.has_job(JobKind.standard)

    def test_launch_privacy_metrics_without_avatar_raises_error(self, launcher):
        """Test that privacy metrics without standard job or avatar data raises error."""
        launcher.config.get_yaml()
        launcher.config.create_privacy_metrics_parameters("test_table", ncp=30)
        set_name = str(uuid4())

        with pytest.raises(
            ValueError,
            match="Expected Avatar tables to be set to run signal/privacy metrics",
        ):
            launcher.launch_job(JobKind.privacy_metrics, set_name)

    def test_launch_privacy_metrics_with_avatar_data(self, launcher):
        """Test launching privacy metrics with avatar data provided."""
        launcher.config.create_table(
            table_name="test_table_with_avatars",
            original_volume="volume",
            original_file="test.csv",
            avatar_volume="volume",
            avatar_file="test_avatars.csv",
        )
        launcher.config.create_privacy_metrics_parameters("test_table_with_avatars", ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.privacy_metrics, set_name)

        assert "privacy_metrics" in launcher.jobs.keys()

    def test_launch_privacy_metrics_after_standard(self, launcher):
        """Test launching privacy metrics after standard job."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)
        launcher.launch_job(JobKind.privacy_metrics, set_name)

        assert "standard" in launcher.jobs.keys()
        assert "privacy_metrics" in launcher.jobs.keys()

    @pytest.mark.parametrize(
        "job_kind",
        [None, JobKind.signal_metrics, JobKind.privacy_metrics],
    )
    def test_launch_report_without_metrics_raises_error(self, job_kind, launcher):
        """Test that launching report without metrics raises error."""
        set_name = str(uuid4())
        if job_kind is not None:
            launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
            launcher.launch_job(JobKind.standard, set_name)
            if job_kind == JobKind.signal_metrics:
                launcher.config.create_signal_metrics_parameters("test_table", ncp=30)
            else:
                launcher.config.create_privacy_metrics_parameters("test_table", ncp=30)
            launcher.launch_job(job_kind, set_name)

        with pytest.raises(
            ValueError,
            match="Expected Privacy and Signal jobs to be created before running report",
        ):
            launcher.launch_job(JobKind.report, set_name)

    def test_launch_report_creates_both_reports(self, launcher):
        """Test that launching report job creates both standard and PIA reports."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        launcher.config.create_privacy_metrics_parameters("test_table", ncp=30)
        launcher.config.create_signal_metrics_parameters("test_table", ncp=30)
        launcher.config.create_report()
        launcher.config.create_report(report_type="pia")
        set_name = str(uuid4())

        # Launch prerequisite jobs
        launcher.launch_job(JobKind.standard, set_name)
        launcher.launch_job(JobKind.privacy_metrics, set_name)
        launcher.launch_job(JobKind.signal_metrics, set_name)

        # Launch report job
        launcher.launch_job(JobKind.report, set_name)

        assert "report" in launcher.jobs
        assert "report_pia" in launcher.jobs
        assert launcher.has_job(JobKind.report)
        assert launcher.has_job("report_pia")

    def test_get_job_dependencies_for_standard(self, launcher):
        """Test that standard job has no dependencies."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)
        dependencies = launcher._get_job_dependencies(JobKind.standard)

        assert dependencies == []

    def test_get_job_dependencies_for_metrics(self, launcher):
        """Test that metrics jobs depend on standard job."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        launcher.config.create_privacy_metrics_parameters("test_table", ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)

        privacy_dependencies = launcher._get_job_dependencies(JobKind.privacy_metrics)
        signal_dependencies = launcher._get_job_dependencies(JobKind.signal_metrics)

        assert len(privacy_dependencies) == 1
        assert len(signal_dependencies) == 1
        assert privacy_dependencies == signal_dependencies

    def test_get_job_dependencies_for_report(self, launcher):
        """Test that report job depends on both metrics jobs."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        launcher.config.create_privacy_metrics_parameters("test_table", ncp=30)
        launcher.config.create_signal_metrics_parameters("test_table", ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)
        launcher.launch_job(JobKind.privacy_metrics, set_name)
        launcher.launch_job(JobKind.signal_metrics, set_name)

        report_dependencies = launcher._get_job_dependencies(JobKind.report)

        assert len(report_dependencies) == 2

    def test_get_job_response_not_found_raises_error(self, launcher):
        """Test that getting non-existent job raises error."""
        with pytest.raises(ValueError, match="Expected job 'standard' to be created"):
            launcher.get_job_response(JobKind.standard)

    def test_has_job_returns_false_when_not_launched(self, launcher):
        """Test has_job returns False for non-launched jobs."""
        assert not launcher.has_job(JobKind.standard)
        assert not launcher.has_job("report_pia")

    def test_has_job_returns_true_when_launched(self, launcher):
        """Test has_job returns True for launched jobs."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)

        assert launcher.has_job(JobKind.standard)
        assert launcher.has_job("standard")

    def test_get_launched_jobs_empty_initially(self, launcher):
        """Test get_launched_jobs returns empty list initially."""
        assert launcher.get_launched_jobs() == []

    def test_get_launched_jobs_returns_all_launched(self, launcher):
        """Test get_launched_jobs returns all launched job names."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        launcher.config.create_privacy_metrics_parameters("test_table", ncp=30)
        launcher.config.create_signal_metrics_parameters("test_table", ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)
        launcher.launch_job(JobKind.privacy_metrics, set_name)
        launcher.launch_job(JobKind.signal_metrics, set_name)

        launched = launcher.get_launched_jobs()

        assert len(launched) == 3
        assert "standard" in launched
        assert "privacy_metrics" in launched
        assert "signal_metrics" in launched

    def test_get_launched_jobs_includes_pia_report(self, launcher):
        """Test that get_launched_jobs includes PIA report when report is launched."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        launcher.config.create_privacy_metrics_parameters("test_table", ncp=30)
        launcher.config.create_signal_metrics_parameters("test_table", ncp=30)
        launcher.config.create_report()
        launcher.config.create_report(report_type="pia")
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)
        launcher.launch_job(JobKind.privacy_metrics, set_name)
        launcher.launch_job(JobKind.signal_metrics, set_name)
        launcher.launch_job(JobKind.report, set_name)

        launched = launcher.get_launched_jobs()

        assert "report" in launched
        assert "report_pia" in launched

    def test_get_job_status(self, launcher):
        """Test getting job status from API."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)
        status = launcher.get_job_status(JobKind.standard)

        assert status.status == "finished"
        assert status.done is True

    def test_get_job_status_by_string(self, launcher):
        """Test getting job status by string name."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        launcher.config.create_privacy_metrics_parameters("test_table", ncp=30)
        launcher.config.create_signal_metrics_parameters("test_table", ncp=30)
        launcher.config.create_report()
        launcher.config.create_report(report_type="pia")
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)
        launcher.launch_job(JobKind.privacy_metrics, set_name)
        launcher.launch_job(JobKind.signal_metrics, set_name)
        launcher.launch_job(JobKind.report, set_name)

        pia_status = launcher.get_job_status("report_pia")

        assert pia_status.status == "finished"
        assert pia_status.done is True

    def test_has_job_with_job_kind_enum(self, launcher):
        """Test has_job works with JobKind enum."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)

        # Should work with enum
        assert launcher.has_job(JobKind.standard)
        # Should also work with string
        assert launcher.has_job("standard")

    def test_get_job_response_with_job_kind_enum(self, launcher):
        """Test get_job_response works with JobKind enum."""
        launcher.config.create_avatarization_parameters("test_table", k=20, ncp=30)
        set_name = str(uuid4())

        launcher.launch_job(JobKind.standard, set_name)

        # Should work with enum
        response_enum = launcher.get_job_response(JobKind.standard)
        # Should also work with string
        response_str = launcher.get_job_response("standard")

        assert response_enum.name == response_str.name



from __future__ import annotations

from pydantic import BaseModel, HttpUrl, model_validator
from typing_extensions import Self

from avatars.config import Config, construct_urls


class ClientConfig(BaseModel):
    """Configuration for ApiClient and Manager.

    This is a plain dataclass-like configuration that can be:
    1. Created directly with explicit values
    2. Derived from a Config object (which loads from env vars)
    3. Passed to ApiClient or Manager

    Attributes
    ----------
    base_url : HttpUrl | None
        The base URL of the Avatar server.
        Default is None.
    base_api_url : HttpUrl | None
        The base URL of the Avatar API server.
        Default is None.
    timeout : int | None
        Timeout in seconds for API requests.
        Default is None (will use DEFAULT_TIMEOUT if not specified).
    should_verify_ssl : bool
        Whether to verify SSL certificates on the server.
        Default is True.
    storage_endpoint_url : HttpUrl | None
        The storage endpoint URL for file uploads/downloads.
        Default is None (must be provided for ApiClient).
    api_key : str | None
        Optional API key for authentication. When provided, this will be used
        instead of username/password authentication.
    """

    base_url: HttpUrl | None = None
    base_api_url: HttpUrl | None = None
    storage_endpoint_url: HttpUrl | None = None

    timeout: int | None = None
    should_verify_ssl: bool = True
    api_key: str | None = None

    @classmethod
    def from_config(cls, config: Config) -> ClientConfig:
        """Convert this Config to a ClientConfig.

        This method extracts the relevant fields from the environment-aware Config
        and creates a ClientConfig that can be passed to ApiClient or Manager.

        Returns
        -------
        ClientConfig
            A new ClientConfig instance with values from this Config.
        """
        # model_construct because we don't want to apply the validators again
        return ClientConfig.model_construct(
            base_url=config.BASE_URL,
            base_api_url=config.BASE_API_URL,
            timeout=config.TIMEOUT,
            should_verify_ssl=config.SHOULD_VERIFY_SSL,
            storage_endpoint_url=config.STORAGE_ENDPOINT_URL,
            api_key=config.API_KEY,
        )

    @model_validator(mode="after")
    def construct_urls_from_base(self) -> Self:
        """Construct API and storage URLs based on BASE_URL if provided.

        This validator runs after model initialization to ensure that
        BASE_API_URL and STORAGE_ENDPOINT_URL are correctly set based
        on BASE_URL if it is provided.

        Returns
        -------
        Self
            The validated Config instance with constructed URLs.
        """
        api_url, storage_url = construct_urls(
            base_url=self.base_url,
            base_api_url=self.base_api_url,
            storage_endpoint_url=self.storage_endpoint_url,
            # We do not want to use defaults here.
            # We shall use from_config to get defaults from Config if needed.
            use_defaults=False,
        )
        self.base_api_url = api_url
        self.storage_endpoint_url = storage_url
        return self



import numpy as np
import pandas as pd
import pandas.testing as pd_testing
import pytest
from pandas.testing import assert_frame_equal

from avatars.processors import InterRecordBoundedRangeDifferenceProcessor


@pytest.fixture
def df() -> pd.DataFrame:
    df = pd.DataFrame(
        {
            "a_s": [30, 100, 80, 70, 40, 70],
            "a_e": [10, 80, 70, 60, 30, 5],
            "b": [4, 3, 0, 0, 2, 4],
            "id": [1, 1, 1, 2, 2, 2],
        }
    )
    return df


def test_process_and_inverse_returns_same_data(df: pd.DataFrame) -> None:
    processor = InterRecordBoundedRangeDifferenceProcessor(
        id_variable="id",
        target_start_variable="a_s",
        target_end_variable="a_e",
        new_first_variable_name="a_s_first_val",
        new_difference_variable_name="a_diff_to_bound",
        new_range_variable="a_range",
        sort_by_variable=None,
        should_round_output=False,
    )

    # Pre-process
    preprocessed_df = processor.preprocess(df)

    # Post-process transformed df
    postprocessed_df = processor.postprocess(df, preprocessed_df)

    pd_testing.assert_frame_equal(postprocessed_df, df, check_dtype=False)


def test_bounds_are_respected(df: pd.DataFrame) -> None:
    processor = InterRecordBoundedRangeDifferenceProcessor(
        id_variable="id",
        target_start_variable="a_s",
        target_end_variable="a_e",
        new_first_variable_name="a_s_first_val",
        new_difference_variable_name="a_diff_to_bound",
        new_range_variable="a_range",
        sort_by_variable=None,
        should_round_output=False,
    )

    # Pre-process
    preprocessed_df = processor.preprocess(df)

    # Post-process shuffled transformed df (this produces values out of bound with other processor)
    shuffled_preprocessed_df = preprocessed_df.sample(frac=1).reset_index(drop=True)
    postprocessed_df = processor.postprocess(df, shuffled_preprocessed_df)

    assert sum(postprocessed_df["a_s"] < 5) == 0  # we should not have any value lower than 5
    assert sum(postprocessed_df["a_e"] < 5) == 0  # we should not have any value lower than 5
    assert sum(postprocessed_df["a_s"] > 100) == 0  # we should not have any value greater than 100
    assert sum(postprocessed_df["a_e"] > 100) == 0  # we should not have any value greater than 100


def test_round_output_produces_integers(df: pd.DataFrame) -> None:
    processor = InterRecordBoundedRangeDifferenceProcessor(
        id_variable="id",
        target_start_variable="a_s",
        target_end_variable="a_e",
        new_first_variable_name="a_s_first_val",
        new_difference_variable_name="a_diff_to_bound",
        new_range_variable="a_range",
        sort_by_variable=None,
        should_round_output=True,
    )

    # Pre-process
    preprocessed_df = processor.preprocess(df)

    # Post-process shuffled transformed df (this produces values out of bound with other processor)
    postprocessed_df = processor.postprocess(df, preprocessed_df)

    assert postprocessed_df["a_s"].dtype in [np.int8, np.int16, np.int32, np.int64]
    assert postprocessed_df["a_e"].dtype in [np.int8, np.int16, np.int32, np.int64]


def test_sort_by_at_preprocess(df: pd.DataFrame) -> None:
    processor = InterRecordBoundedRangeDifferenceProcessor(
        id_variable="id",
        target_start_variable="a_s",
        target_end_variable="a_e",
        new_first_variable_name="a_s_first_val",
        new_difference_variable_name="a_diff_to_bound",
        new_range_variable="a_range",
        sort_by_variable="b",
        should_round_output=False,
    )

    # Pre-process
    preprocessed_df = processor.preprocess(df)

    expected = pd.DataFrame(
        {
            "b": [4, 3, 0, 0, 2, 4],
            "id": [1, 1, 1, 2, 2, 2],
            "a_range": [-0.8, -0.211, -0.133, -0.154, -0.286, -1.0],
            "a_s_first_val": [80, 80, 80, 70, 70, 70],
            "a_diff_to_bound": [-0.666, 1.0, 0.0, 0.0, -0.364, 0.571],
        }
    )

    assert_frame_equal(preprocessed_df, expected, rtol=0.01)


def test_preprocess_raises_error_when_wrong_variable_name(df: pd.DataFrame) -> None:
    processor = InterRecordBoundedRangeDifferenceProcessor(
        id_variable="wrong_id",
        target_start_variable="a_s",
        target_end_variable="a_e",
        new_first_variable_name="a_s_first_val",
        new_difference_variable_name="a_diff_to_bound",
        new_range_variable="a_range",
    )

    with pytest.raises(
        ValueError,
        match="Expected valid variable names for",
    ):
        processor.preprocess(df)


def test_preprocess_raises_error_when_missing_values(df: pd.DataFrame) -> None:
    df.loc[1, "a_s"] = np.nan
    processor = InterRecordBoundedRangeDifferenceProcessor(
        id_variable="id",
        target_start_variable="a_s",
        target_end_variable="a_e",
        new_first_variable_name="a_s_first_val",
        new_difference_variable_name="a_diff_to_bound",
        new_range_variable="a_range",
    )

    with pytest.raises(
        ValueError,
        match="Expected no missing values for",
    ):
        processor.preprocess(df)


def test_extra_columns_are_kept_at_postprocess(df: pd.DataFrame) -> None:
    processor = InterRecordBoundedRangeDifferenceProcessor(
        id_variable="id",
        target_start_variable="a_s",
        target_end_variable="a_e",
        new_first_variable_name="a_s_first_val",
        new_difference_variable_name="a_diff_to_bound",
        new_range_variable="a_range",
        sort_by_variable=None,
        should_round_output=False,
    )

    # Pre-process
    preprocessed_df = processor.preprocess(df)

    # Add an extra variable (this would typically happen when preprocessing many variables)
    preprocessed_df["extra_variable_1"] = preprocessed_df["a_range"]
    df["extra_variable_2"] = df["a_s"]

    postprocessed_df = processor.postprocess(df, preprocessed_df)

    # Post-process df should have the additional variables from dest but not the additional
    # variables from source
    should_have_columns = df.columns.to_list() + ["extra_variable_1"]
    should_have_columns.remove("extra_variable_2")

    assert postprocessed_df.columns.tolist() == should_have_columns



"""Tests for ClientConfig model_validator."""

import pytest
from pydantic import HttpUrl

from avatars.client_config import ClientConfig
from avatars.config import Config


class TestClientConfigValidator:
    """Test the construct_urls_from_base model validator."""

    def test_base_url_constructs_api_and_storage_urls(self) -> None:
        """Test that base_url constructs both API and storage URLs."""
        config = ClientConfig(base_url="https://example.com")

        assert str(config.base_api_url) == "https://example.com/api"
        assert str(config.storage_endpoint_url) == "https://example.com/storage"

    def test_base_url_and_explicit_api_url_raises(self) -> None:
        with pytest.raises(ValueError, match="Cannot set BASE_URL together with"):
            ClientConfig(
                base_url="https://example.com",
                base_api_url="https://api.example.com",
            )

    def test_base_url_and_explicit_storage_url_raises(self) -> None:
        with pytest.raises(ValueError, match="Cannot set BASE_URL together with"):
            ClientConfig(
                base_url="https://example.com",
                storage_endpoint_url="https://storage.example.com",
            )

    def test_base_url_and_explicit_storage_and_api_url_raises(self) -> None:
        with pytest.raises(ValueError, match="Cannot set BASE_URL together with"):
            ClientConfig(
                base_url="https://example.com",
                base_api_url="https://api.example.com",
                storage_endpoint_url="https://storage.example.com",
            )

    def test_explicit_urls_without_base_url(self) -> None:
        """Test that explicit API and storage URLs work without base_url."""
        config = ClientConfig(
            base_api_url="https://api.example.com",
            storage_endpoint_url="https://storage.example.com",
        )

        assert str(config.base_api_url) == "https://api.example.com/"
        assert str(config.storage_endpoint_url) == "https://storage.example.com/"
        assert config.base_url is None

    def test_from_config_method(self) -> None:
        """Test that from_config properly converts Config to ClientConfig."""
        # Don't use BASE_URL with Config because it reads from env and may conflict
        env_config = Config(
            _env_file=None,  # Don't load from .env
            BASE_API_URL=HttpUrl("https://api.example.com"),
            STORAGE_ENDPOINT_URL=HttpUrl("https://storage.example.com"),
            TIMEOUT=90,
            SHOULD_VERIFY_SSL=False,
            API_KEY="test-key",
        )

        client_config = ClientConfig.from_config(env_config)

        assert str(client_config.base_api_url) == "https://api.example.com/"
        assert str(client_config.storage_endpoint_url) == "https://storage.example.com/"
        assert client_config.timeout == 90
        assert client_config.should_verify_ssl is False
        assert client_config.api_key == "test-key"



import pandas as pd

from avatars.lib.continuous_threshold import get_continuous_under_threshold


def test_get_continuous_under_threshold() -> None:
    """Check the variable selection with a threshold."""
    df = pd.DataFrame({"variable_1": [1, 1, 2], "variable_2": [1, 2, 3]})

    variables = get_continuous_under_threshold(df, threshold=2)
    expected = ["variable_1"]

    assert variables == expected



from datetime import datetime

import pandas as pd
import pytest


@pytest.fixture
def many_dtypes_df() -> pd.DataFrame:
    df = pd.DataFrame()
    df["ints"] = [-2, 0, 2, 3, 4, 5, 6]
    df["floats"] = [-2.1, 0.0, 2.1, 3.1, 4.1, 5.1, 6.1]
    df["strings"] = ["Mr", "John", "Doe", "Mrs.", "Mary", "Higgins", "Clark"]
    df["datetimes"] = pd.to_datetime([datetime(2022, 1, 1).replace(tzinfo=None)] * 7)
    return df


@pytest.fixture
def categorical_df() -> pd.DataFrame:
    df = pd.DataFrame(
        {
            "variable_1": ["red", "blue", "blue", "green"],
            "variable_2": ["red", "blue", "blue", "red"],
            "variable_3": ["green", "green", "green", "green"],
        }
    )
    return df
